@@ -85,6 +85,13 @@ class PlotsItselfMixin:
 
                 plt.scatter(xs, ys, marker='.', c=palette[i % len(palette)], s=2)
 
+    def plot_each_class(self, **kwargs):
+        """Plot each class on its own"""
+        for yi in sorted(list(set(self.y))):
+            mask = self.y == yi
+            title = self.classmap.get(yi, "Class #%d" % yi)
+            self.replace().mask(mask).plot(title=title, **kwargs)
+
     def plot_class_distribution(self, force=False):
         """
         Plot histogram of classes' samples
@@ -119,6 +126,7 @@ class PlotsItselfMixin:
             return
 
         y_segments = self.y_segments()
+        results = []
 
         for class_idx in set(self.y):
             if classes is not None and class_idx not in classes:
@@ -145,7 +153,13 @@ class PlotsItselfMixin:
 
             hist, edges = np.histogram(durations, bins=bins)
 
-            return np.cumsum(hist), edges[:-1]
+            results.append({
+                "class": class_idx,
+                "cumsum": np.cumsum(hist),
+                "edges": edges[:-1]
+            })
+
+        return results
 
     def plot_pairplot(self, max_samples=500, force=False, palette=None, **kwargs):
         """
